/*
  LyraCalc
  Arduino C firmware for the Lyratron Pseudoscientific Calculator
  (a SparkFun Arduino Pro mated to a Lyratron Calc Shield)
  (cc) BY-NC-SA Peter Alaric DeSimone
  http://www.lyratron.com
*/

// LCD commands
#define LCD_COMMAND  254
#define LCD_CONTROL  0x7C

// scan codes
#define ONE_KEY      61
#define TWO_KEY      51
#define THREE_KEY    41
#define FOUR_KEY     59
#define FIVE_KEY     49
#define SIX_KEY      39
#define SEVEN_KEY    57
#define EIGHT_KEY    47
#define NINE_KEY     37
#define ZERO_KEY     45
#define ADD_KEY      31
#define SUB_KEY      21
#define MUL_KEY      29
#define DIV_KEY      19
#define POINT_KEY    55
#define SQUARE_KEY   17
#define NPOWER_KEY   15
#define NROOT_KEY    25
#define FACT_KEY     23
#define SQRT_KEY     27
#define NROOT        25
#define CLEAR_KEY    35
#define MPLUS_KEY    53
#define MR_KEY       43
#define MC_KEY       33
#define EQUALS_KEY   13

// commands
#define ADD_COMMAND       10
#define SUB_COMMAND       20
#define MUL_COMMAND       30
#define DIV_COMMAND       40
#define SQUARE_COMMAND    50
#define SQRT_COMMAND      60
#define NPOWER_COMMAND    70
#define NROOT_COMMAND     80
#define FACT_COMMAND      90
#define MPLUS_COMMAND     100
#define MR_COMMAND        110
#define MC_COMMAND        120
#define CLEAR_COMMAND     130
#define EQUALS_COMMAND    140

#define STRING_LENGTH  100

int keyPressed = 0;
char currentString[STRING_LENGTH];
int currentCommand = 0;
int strIndex = 0;

float M = 0;

String line1, line2;

// first and second numbers: these are the numbers inputted into the
// calculator that it will perform operations on.  the first number
// is always inputted first, and the second, well... you get the idea.
// in the case of "2 * 3", '2' would be the first number, '3' the second
// and '*' (MULTIPLY) the operator.  '2' is stored as firstNumber
// immediately when any operator key (in this case, '*') is pressed.
// similarly, secondNumber is stored when any operator key (including '=')
// is pressed.  pressing '=' immediately after inputting firstNumber will
// merely parrot back firstNumber, following by the message ("...duh!").
float firstNumber=0,
      secondNumber=0;


void setup()
{
  Serial.begin(9600); // for SparkFun 16x2 serial LCD display
  LCDreset();
  checkBattery();
  menu();
}


void loop()
{
  calculator();
}


void checkBattery()
{
  float batteryVoltage = analogRead(0) * 4.88; //(analogRead(0) / 1023) * 5;
  LCDclear();
  LCDline1();
  Serial.print("batt: ");
  Serial.print(batteryVoltage);
  Serial.print("mV");
  LCDline2();
  float percentage = (batteryVoltage / 1500) * 100;
  Serial.print("(");
  Serial.print(percentage);
  Serial.print("%)");
  delay(1000);
}


void menu()
{
  LCDclear();
  LCDline1();
  Serial.print("1:calc 2:solve");
  LCDline2();
  Serial.print("lin 3:solve quad");
  int k = getch();
  while (k != ONE_KEY &&
         k != TWO_KEY &&
         k != THREE_KEY)
           k = getch();
  if (k == TWO_KEY) solveLinearLoop();
  else if (k == THREE_KEY) solveQuadLoop();
  // default is to continue on to calc
}


void solveLinearLoop()
{
  while (1) solveLinear();
}


void solveLinear()
{
  // a
  LCDclear();
  LCDline1();
  Serial.print("ax + b = cx + d");
  LCDline2();
  Serial.print("a=");
  inputString();
  float a = parseString();
  // b
  LCDclear();
  LCDline1();
  Serial.print("ax + b = cx + d");
  LCDline2();
  Serial.print("b=");
  inputString();
  float b = parseString();
  // c
  LCDclear();
  LCDline1();
  Serial.print("ax + b = cx + d");
  LCDline2();
  Serial.print("c=");
  inputString();
  float c = parseString();
  // d
  LCDclear();
  LCDline1();
  Serial.print("ax + b = cx + d");
  LCDline2();
  Serial.print("d=");
  inputString();
  float d = parseString();
  // now solve for x
  LCDclear();
  float x = (d - b) / (a - c);
  LCDline1();
  Serial.print(a);
  Serial.print(",");
  Serial.print(b);
  Serial.print(",");
  Serial.print(c);
  Serial.print(",");
  Serial.print(d);
  LCDline2();
  Serial.print("x=");
  Serial.print(x);
  getch();
}


void solveQuadLoop()
{
  while (1) solveQuad();
}


void solveQuad()
{
  // a
  LCDclear();
  LCDline1();
  Serial.print("ax^2+bx+c=0");
  LCDline2();
  Serial.print("a=");
  inputString();
  float a = parseString();
  // b
  LCDclear();
  LCDline1();
  Serial.print("ax^2+bx+c=0");
  LCDline2();
  Serial.print("b=");
  inputString();
  float b = parseString();
  // c
  LCDclear();
  LCDline1();
  Serial.print("ax^2+bx+c=0");
  LCDline2();
  Serial.print("c=");
  inputString();
  float c = parseString();
  // now solve for x using the quadratic formula
  LCDclear();
  float bSquared = pow(b, 2);
  float ac4 = 4 * a * c;
  float underRadical = bSquared - ac4;
  float a2 = 2 * a;
  float numerator = sqrt(underRadical) - b;
  float x = numerator / a2;
  LCDline1();
  Serial.print(a);
  Serial.print(",");
  Serial.print(b);
  Serial.print(",");
  Serial.print(c);
  LCDline2();
  Serial.print("x=");
  Serial.print(x);
  getch();
}


int calculator()
{
  LCDclear();
  
  currentCommand = 0;
  float answer = 0;
  
  LCDline1();
  inputString();
  float n1 = parseString();

  // check for "clear" and "1-variable" commands we can execute immediately
  if (currentCommand == CLEAR_COMMAND)
  {
    LCDclear();
    return 0;
  }
  else if (currentCommand == SQUARE_COMMAND)
  {
    answer = pow(n1, 2);
  }
  else if (currentCommand == SQRT_COMMAND)
  {
    answer = sqrt(n1);
  }
  else if (currentCommand == FACT_COMMAND)
  {
    answer = factorial(n1);
  }
  else if (currentCommand == MPLUS_COMMAND)
  {
    M += n1;
    return 0;
  }
  else if (currentCommand == MC_COMMAND)
  {
    M = 0;
    return 0;
  }
  else
  {
    // it's a 2-variable command
    
    LCDline2();
    inputString();
    float n2 = parseString();
    
    LCDclear();
    
    LCDline1();
    Serial.print(n1);

    if (currentCommand == ADD_COMMAND)
    {
      answer = n1 + n2;
      Serial.print("+");
    }
    else if (currentCommand == SUB_COMMAND)
    {
      answer = n1 - n2;
      Serial.print("-");
    }
    if (currentCommand == MUL_COMMAND)
    {
      answer = n1 * n2;
      Serial.print("*");
    }
    else if (currentCommand == DIV_COMMAND)
    {
      answer = n1 / n2;
      Serial.print("/");
    }
    else if (currentCommand == NPOWER_COMMAND)
    {
      answer = pow(n1, n2);
      Serial.print("^");
    }
    else if (currentCommand == NROOT_COMMAND)
    {
      answer = pow(n1, (1/n2));
      Serial.print("^1/");
    };
  
    Serial.print(n2);
  }
  
  Serial.print("=");
  
  LCDline2();
  Serial.print(answer);

  if (currentCommand == CLEAR_COMMAND)
  {
    LCDclear();
    return 0;
  };
  
  waitForKey();
}


float factorial(float x)
{
  float f = 0;
  if (isInt(x))
  {
    f = x;
    for (; x >= 2; x--)
    {
      f = f * (x - 1);
    };
  }
  else
  {
    // can't factorial a float!
  };
  return f;
}


int isInt(float f)
{
  // checks to see if a number is an integer or not
  if (f == int(f))
    return 1;
  else
    return 0;
}


float parseString()
{
  float n = 0;
  if (currentCommand == MR_COMMAND)
  {
    n = M;
    return n;
  };
  int multiplier = 1;
  int decimalPoint = 0;
  int digitsRightOfDecimal = 0;
  int negative = 0;
  for (int i = 0; i < STRING_LENGTH; i++)
  {
    char c = currentString[i];
    if (c == '\0') break;
    int d = 0;
    if (c == '.') decimalPoint++;
    else if (c == '-') negative++;
    else
    {
      if (c == '0') d = 0;
      else if (c == '1') d = 1;
      else if (c == '2') d = 2;
      else if (c == '3') d = 3;
      else if (c == '4') d = 4;
      else if (c == '5') d = 5;
      else if (c == '6') d = 6;
      else if (c == '7') d = 7;
      else if (c == '8') d = 8;
      else if (c == '9') d = 9;
      n *= 10;
      n += d;
      if (decimalPoint) digitsRightOfDecimal++;
    };
  };
  n /= pow(10, digitsRightOfDecimal);
  if (negative) n = -n;
  return n;
}


void inputString()
{
  int k = getch();
  int i = 0;
  while (k == ONE_KEY ||
         k == TWO_KEY ||
         k == THREE_KEY ||
         k == FOUR_KEY ||
         k == FIVE_KEY ||
         k == SIX_KEY ||
         k == SEVEN_KEY ||
         k == EIGHT_KEY ||
         k == NINE_KEY ||
         k == ZERO_KEY ||
         k == POINT_KEY ||
         (k == SUB_KEY && i == 0))
  {
    char c = 0;
    switch(k)
    {
      // numbers
      case ONE_KEY:
        c = '1';
        break;
      case TWO_KEY:
        c = '2';
        break;
      case THREE_KEY:
        c = '3';
        break;
      case FOUR_KEY:
        c = '4';
        break;
      case FIVE_KEY:
        c = '5';
        break;
      case SIX_KEY:
        c = '6';
        break;
      case SEVEN_KEY:
        c = '7';
        break;
      case EIGHT_KEY:
        c = '8';
        break;
      case NINE_KEY:
        c = '9';
        break;
      case ZERO_KEY:
        c = '0';
        break;
      case POINT_KEY:
        c = '.';
        break;
      case SUB_KEY: // only for leading negative sign!
        c = '-';
        break;
      default:
        c = '\0';
    };
    Serial.print(c);
    currentString[i] = c;
    i++;
    currentString[i] = '\0';
    k = getch();
  };
  saveCurrentCommand();
}

void saveCurrentCommand()
{
  int k = keyPressed;
  switch (k)
  {
      // commands
      case ADD_KEY:
        currentCommand = ADD_COMMAND;
        Serial.print("+");
        break;
      case SUB_KEY:
        currentCommand = SUB_COMMAND;
        Serial.print("-");
        break;
      case MUL_KEY:
        currentCommand = MUL_COMMAND;
        Serial.print("*");
        break;
      case DIV_KEY:
        currentCommand = DIV_COMMAND;
        Serial.print("/");
        break;
      case SQUARE_KEY:
        currentCommand = SQUARE_COMMAND;
        Serial.print("^2");
        break;
      case SQRT_KEY:
        currentCommand = SQRT_COMMAND;
        Serial.print("[sqrt]");
        break;
      case NPOWER_KEY:
        currentCommand = NPOWER_COMMAND;
        Serial.print("^");
        break;
      case NROOT_KEY:
        currentCommand = NROOT_COMMAND;
        Serial.print("[nRT] n=");
        break;
      case FACT_KEY:
        currentCommand = FACT_COMMAND;
        Serial.print("!");
        break;
      case MPLUS_KEY:
        currentCommand = MPLUS_COMMAND;
        Serial.print("M+");
        break;
      case MR_KEY:
        currentCommand = MR_COMMAND;
        Serial.print("MR");
        break;
      case MC_KEY:
        currentCommand = MC_COMMAND;
        Serial.print("MC");
        break;
      case CLEAR_KEY:
        currentCommand = CLEAR_COMMAND;
        Serial.print("CLEAR");
        break;
      case EQUALS_KEY:
        //currentCommand = EQUALS_COMMAND; // don't erase the current command - EQUALS isn't really a command!
        break;
  };
}


int getch()
{
  waitForKey();
  int k = keyPressed;
  tone(13, 1000, 50);
  waitForRelease();
  delay(100); // to prevent jitter/bounce
  keyPressed = k;
  /*
  LCDline1();
  Serial.print("code=");
  Serial.print(k);
  */
  return k;
}

void outputCurrentScanCode()
{
  
  Serial.print(keyPressed, DEC);
}

void waitForKey()
{
  scanKeypad();
  while (keyPressed == 0) scanKeypad();
}
void waitForRelease()
{
  scanKeypad();
  while (keyPressed != 0) scanKeypad();
}


// WARNING: ALTERING THE FOLLOWING CODE COULD BE VERY DANGEROUS
// FOR YOUR ARDUINO!  the Lyratron Calc Shield does not use
// resistors to protect I/O pins from damage due to overcurrent.
// when you press a button on your calculator, two I/O pins are
// instantly connected to each other.  this is not a problem if
// at least one of the two pins is in INPUT ("high impedance")
// mode.  but if both pins are in OUPUT ("low impedance") mode,
// a huge amount of current will flow between them, destroying
// your Arduino!  so our advice is: don't mess with this code!
// (there really isn't anything worth changing in this part anyway).
void scanKeypad()
{
  keyPressed = 0;
  allYpinsHigh(); // set all 'y' pins to HIGH; set low one at a time to scan
  for (int y = 2; y <= 12; y+=2) // y = even pins; 2, 4, 6, 8, 10
  {
    // y is our send pin and x is our receive pin
    // send a LOW on this row
    pinMode(y, OUTPUT);
    digitalWrite(y, LOW);
    for (int x = 3; x <= 11; x+=2) // x = odd pins; 3, 5, 7, 9, 11
    {
      // scan this key
      pinMode(x, INPUT);
      digitalWrite(x, HIGH); // pulled high in INPUT MODE ONLY!!!
      // y is our send pin and x is our receive pin
      // read the x pin to check if connected to the current y
      // if it is, y will pull it LOW against x's internal pullup resistor.
      // otherwise, it will drift HIGH.
      /*
      LCDclear();
      LCDline1();
      Serial.print("x");
      Serial.print(x);
      Serial.print("y");
      Serial.print(y);
      Serial.print("=");
      */
      if (digitalRead(x) == LOW)
      {
        //Serial.print("0");
        // flag this key as being pressed
        keyPressed = 5 * y + x;
      }
      else
      {
        //Serial.print("1");
      };
      //delay(25);
    };
    // we're done checking this row
    // switch this row back to being HIGH
    digitalWrite(y, HIGH);
  };
}

void allYpinsHigh()
{
  for (int y = 2; y <= 12; y+=2)
  {
    pinMode(y, OUTPUT);
    digitalWrite(y, HIGH);
  };
}




void LCDout(String s)
{
  line1 = line2;
  line2 = s;
  LCDline1();
  Serial.print(line1);
  LCDline2();
  Serial.print(line2);
}


void LCDreset()
{
  Serial.write(LCD_CONTROL); // max brightness
  Serial.write("r"); // reset
}


void LCDclear()
{
  Serial.write(LCD_COMMAND);
  Serial.write(0x01);
}


void LCDline1()
{
  Serial.write(LCD_COMMAND); // 1st line
  Serial.write(0x80 + 0);
}


void LCDline2()
{
  Serial.write(LCD_COMMAND); // 2nd line
  Serial.write(0x80 + 64);
}
